# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2024/12/18 下午7:01
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

from dataclasses import dataclass

class WarnException(Exception):
    """自定义异常类，仅用于警告"""
    def __init__(self, message):
        super().__init__(message)  # 调用父类的构造函数

@dataclass
class FailTaskError:
    task_name: str
    error: Exception

    def __str__(self):
        return f"""
[失败任务]: {self.task_name}
[错误信息]: \n{self.error}
"""

    def __repr__(self):
        return self.__str__()


# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/7/20 16:33
# Description:
import queue

# from ygo.utils import locate
from ygo.pool import run_job

# def test_locate():
    # fn = locate("functools.partial")
    # test_fn = lambda a, b: a+b
    # assert fn(test_fn, 1, 2)() == 3

def foo():
    run_job(1, 1, queue.Queue())

if __name__ == '__main__':
    foo()


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/5/27 09:14
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import time

import ygo


def sample_task_add(a, b):
    time.sleep(1)
    return a + b


def sample_task_sub(a, b):
    time.sleep(1)
    return a - b


def warn_task():
    raise ygo.WarnException("Test warning")


def error_task():
    raise ValueError("Test error")


def test_pool_submit_and_do():
    with ygo.pool() as go:
        go.submit(sample_task_add, "task add")(a=1, b=2)
        go.submit(sample_task_add, "task add")(a=2, b=3)
        go.submit(sample_task_sub, "task sub")(a=5, b=1)

        results = go.do()

    assert len(results) == 3
    assert results[0] == 3
    assert results[1] == 5
    assert results[2] == 4


def test_pool_error_handling():
    with ygo.pool() as go:
        go.submit(error_task, job_name="task_error")()

        results = go.do()

    assert len(results) == 1
    assert results[0] is None


def test_pool_warn_exception_handled():
    """测试 WarnException 是否被记录警告但继续执行"""
    with ygo.pool(n_jobs=1) as go:
        go.submit(warn_task, job_name="task_warn")()

        results = go.do()

        assert len(results) == 1
        assert results[0] is None  # WarnException 不中断流程


def test_pool_progress_bar_disabled():
    """测试关闭进度条时的行为"""
    with ygo.pool(n_jobs=2, show_progress=False) as go:
        go.submit(sample_task_add, "test add")(a=1, b=2)
        go.submit(sample_task_add, "test add")(a=3, b=4)

        results = go.do()

        assert len(results) == 2
        assert 3 == results[0]
        assert 7 == results[1]


def test_pool_with_one_jobs():
    """测试 n_jobs=1 时是否退化为串行执行"""
    with ygo.pool(n_jobs=1) as go:
        go.submit(sample_task_add, "test add")(a=1, b=2)
        go.submit(sample_task_add, "test add")(a=3, b=4)

        results = go.do()

        assert len(results) == 2
        assert 3 == results[0]
        assert 7 == results[1]


def test_delay_basic_usage():
    """场景1：基本使用"""
    fn = ygo.delay(lambda a, b: a + b)(a=1, b=2)
    assert fn() == 3


def test_delay_partial_kwargs():
    """场景2：逐步传递参数"""
    fn1 = ygo.delay(lambda a, b, c: a + b + c)(a=1)
    fn2 = ygo.delay(fn1)(b=2)
    assert fn2(c=3) == 6


def test_delay_override_kwargs():
    """场景3：参数更改"""
    fn1 = ygo.delay(lambda a, b, c: a + b + c)(a=1, b=2)
    fn2 = ygo.delay(fn1)(c=3, b=5)  # 修改 b 的值
    assert fn2() == 9


def test_delay_with_no_call():
    """延迟函数在未调用时不会执行"""
    called = False

    def side_effect(*args, **kwargs):
        nonlocal called
        called = True
        return None

    delayed_fn = ygo.delay(side_effect)(x=1)
    assert not called  # 还未执行
    delayed_fn()
    assert called  # 执行后标记为 True


def test_fn_signature_params():
    assert ygo.fn_signature_params(sample_task_add) == ['a', 'b']


def test_fn_params_with_defaults():
    delayed = ygo.delay(sample_task_add)(a=3)
    print(ygo.fn_params(delayed))
    assert dict(ygo.fn_params(delayed)) == {'a': 3, }


def test_fn_path_for_function():
    fn_path = ygo.fn_path(sample_task_add)
    assert 'tests.test_ygo' in fn_path or '__main__' in fn_path


def test_fn_code():
    code = ygo.fn_code(sample_task_add)
    assert "def sample_task_add(a, b):" in code
    assert "return a+b" in code


def test_fn_info():
    info = ygo.fn_info(sample_task_add)
    assert "sample_task_add(a, b)" in info
    assert "def sample_task_add(a, b):" in info


def test_fn_from_str():
    func = ygo.fn_from_str("ygo.utils.fn_from_str")
    assert func.__name__ == "fn_from_str"


def test_module_from_str():
    mod = ygo.module_from_str("ygo.utils")
    assert mod.__name__ == "ygo.utils"


def test_fn_params_with_no_defaults():
    def no_defaults(x, y):
        pass

    delayed = ygo.delay(no_defaults)()
    assert dict(ygo.fn_params(delayed)) == {}



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2024/11/4 下午2:10
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
import functools
import multiprocessing
import threading
import warnings

from joblib import Parallel, delayed
import logair
from .exceptions import WarnException, FailTaskError
from .delay import delay

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from tqdm.auto import tqdm

logger = logair.get_logger("ygo")

def run_job(job, task_id, queue):
    """执行任务并更新队列"""
    try:
        result = job()
    except WarnException as e:
        logger.warning(FailTaskError(task_name=job.task_name, error=e))
        result = None
    except Exception as e:
        logger.error(FailTaskError(task_name=job.task_name, error=e), exc_info=e)
        result = None
    queue.put((task_id, 1))
    return result


def update_progress_bars(tqdm_objects: list[tqdm],
                         task_ids,
                         queue: multiprocessing.Queue,
                         num_tasks: int,
                         task_counts: dict):
    """根据队列中的消息更新 tqdm 进度条"""
    completed_tasks = 0
    completed_task_jobs = {id_: 0 for id_ in task_ids}
    while completed_tasks < num_tasks:
        task_id, progress_value = queue.get()  # 从队列获取进度更新
        completed_task_jobs[task_id] += 1
        if completed_task_jobs[task_id] >= task_counts[task_id]:
            completed_tasks += 1
        tqdm_objects[task_id].update(progress_value)
    [tqdm_object.close() for tqdm_object in tqdm_objects]


class pool:
    """
    每个fn运行一次算一个job，每个job需要指定job_name, 如果没有job_name, 则默认分配 TaskDefault
    相同 job_name 的fn归到同一个task, 同时该task命名为job_name
    即一个task中包含了多个需要运行的 job fn
    task1 <job_fn1, job_fn2, ...>
    task2 <job_fn3, job_fn4, ...>
    所有的job_fn都会通过joblib并行运行
    """

    def __init__(self, n_jobs=5, show_progress=True, backend='threading', ):
        """backend: loky/threading/multiprocessing"""
        self.show_progress = show_progress
        self._n_jobs = n_jobs
        self._parallel = Parallel(n_jobs=self._n_jobs, verbose=0, backend=backend) if self._n_jobs > 1 else None
        self._default_task_name = "GO-JOB"
        self._all_jobs = list()  # list[job]
        self._all_tasks = list()  # list[task_name]
        self._task_ids = dict()  # {task_name1: 0, task_name2: 1, ...}
        self._task_counts = dict()
        self._id_counts = dict()
        self._manager = None

    def submit(self, fn, job_name=""):
        """
        提交并行任务
        Parameters
        ----------
        fn: callable
            需要并行的callable对象
        job_name: str
            提交到的任务名, 不同的任务对应不同的进度条
        Returns
        -------

        Examples
        --------
        import time
        import ygo
        >>> def task_fn1(a, b):
                time.sleep(3)
                return a+b
        >>> def task_fn2():
                time.sleep(5)
                return 0
        >>> with ygo.pool() as go:
                go.submit(task_fn1, job_name="task1")(a=1, b=2)
                go.submit(task_fn2, job_name="task2")()

                go.do()
        """

        # 提交任务，对任务进行分类，提交到对应的task id中，并且封装新的功能：使其在运行完毕后将任务进度更新放入队列
        @functools.wraps(fn)
        def collect(**kwargs):
            """归集所有的job到对应的task"""
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                job = delay(fn)(**kwargs)
                task_name = self._default_task_name if not job_name else job_name
                if task_name not in self._task_ids:
                    self._task_ids[task_name] = len(self._all_tasks)
                    self._task_counts[task_name] = 0
                    self._all_tasks.append(task_name)
                    self._id_counts[self._task_ids[task_name]] = 0
                self._task_counts[task_name] += 1
                self._id_counts[self._task_ids[task_name]] += 1
                job.task_id = self._task_ids[task_name]
                job.job_id = self._task_counts[task_name]
                job.task_name = task_name
                self._all_jobs.append(job)
                return job

        return collect

    def do(self):
        if self.show_progress:
            # 消息队列进行通信
            self._manager = multiprocessing.Manager()
            queue = self._manager.Queue()
            tqdm_bars = [tqdm(total=self._task_counts[task_name],
                              desc=f"{str(task_name)}", leave=False) for task_name in
                         self._all_tasks]
            # 初始化多个任务的进度条，每个任务一个
            task_ids = [task_id for task_id in range(len(self._all_tasks))]
            # 创建并启动用于更新进度条的线程
            progress_thread = threading.Thread(target=update_progress_bars, args=(
                tqdm_bars, task_ids, queue, len(self._all_tasks), self._id_counts))
            progress_thread.start()
            if self._parallel is not None:
                # 执行并行任务
                result = self._parallel(delayed(run_job)(job=job,
                                                         task_id=job.task_id,
                                                         queue=queue) for job in self._all_jobs)
            else:
                result = [run_job(job=job, task_id=job.task_id, queue=queue) for job in self._all_jobs]
            # 等待进度更新线程执行完毕
            progress_thread.join()
        else:
            if self._parallel is not None:
                result = self._parallel(delayed(job)() for job in self._all_jobs)
            else:
                result = [job() for job in self._all_jobs]
        self._all_jobs = list()  # list[job]
        self._all_tasks = list()  # list[task_name]
        self._task_ids = dict()  # {task_name1: 0, task_name2: 1, ...}
        self._task_counts = dict()
        self._id_counts = dict()
        self.close()
        return result

    def close(self):
        """释放所有资源"""
        if hasattr(self, '_parallel') and self._parallel is not None:
            try:
                self._parallel.__exit__(None, None, None)
            except Exception as e:
                logger.warning(f"Failed to close Parallel: {e}")

        if hasattr(self, '_manager') and self._manager is not None:
            try:
                self._manager.shutdown()
            except Exception as e:
                logger.warning(f"Failed to shutdown Manager: {e}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # 释放进程
        self.close()
